import pandas as pd
import glob
from collections.abc import Iterable
import numpy as np


def load_phylo(filename):
    df = pd.read_csv(filename)
    return df


def assign_relative_fit(phylolog):
    phylolog.columns = [col.replace(" ", "") for col in phylolog.columns]
    # print(phylolog.columns)
    # get original function values (multiple in some cases)
    original_fitness = phylolog.at[0, "hash_fitness"]
    # remove original function from frame
    phylolog.drop(0, axis=0)
    # add relative fitness (with group by to match metrics)
    phylolog["rel_fitness"] = phylolog.hash_fitness / original_fitness
    return phylolog


def save_phylo(phylolog, filename):
    # save phylolog as csv with updated name
    new_file = filename[:-12] + "rankedphylolog.csv"
    phylolog.to_csv(new_file)
    pass


def assign_rank(relative_phylo):
    relative_phylo["rank"] = relative_phylo.groupby("generation")["rel_fitness"].rank("dense", ascending=True)
    return relative_phylo


def assign_rank_change(ranked_phylo):
    # acquire rank of ancestor
    ranked_phylo["ancestor_rank"] = ranked_phylo.apply(
        lambda x: get_ancestor_rank(x["hash_id"], x["hash_ancestorsList"], x["generation"], ranked_phylo), axis=1)
    # calculate difference in rank
    ranked_phylo["rank_change"] = ranked_phylo["rank"]-ranked_phylo["ancestor_rank"]
    return ranked_phylo


def assign_relative_fit_change(ranked_phylo):
    # acquire rank of ancestor
    ranked_phylo["ancestor_rfit"] = ranked_phylo.apply(
        lambda x: get_ancestor_rel_fit(x["hash_id"], x["hash_ancestorsList"], x["generation"], ranked_phylo), axis=1)
    # calculate difference in rank
    ranked_phylo["rfit_change"] = ranked_phylo["rel_fitness"]-ranked_phylo["ancestor_rfit"]
    return ranked_phylo


def get_ancestor_rel_fit(child_id, ancestor, generation, phylolog):
    ancestor = int(ancestor[2:-1])
    # check if ancestor is in previous generation
    if generation == 0:
        ancestor_record = phylolog.loc[(phylolog["hash_id"] == ancestor) & (phylolog["generation"] == 0)]
    else:
        ancestor_record = phylolog.loc[(phylolog["hash_id"] == ancestor) & (phylolog["generation"] == generation-1)]
    if not ancestor_record.empty:
        # if ancestor is in previous generation return rank
        return ancestor_record.iloc[0]["rel_fitness"]
    # else return ids rank from previous generation
    else:
        prev_gen = phylolog.loc[(phylolog["hash_id"] == child_id) & (phylolog["generation"] == generation)]
        return prev_gen.iloc[0]["rel_fitness"]


def get_ancestor_rank(child_id, ancestor, generation, phylolog):
    ancestor = int(ancestor[2:-1])
    # check if ancestor is in previous generation
    if generation == 0:
        ancestor_record = phylolog.loc[(phylolog["hash_id"] == ancestor) & (phylolog["generation"] == 0)]
    else:
        ancestor_record = phylolog.loc[(phylolog["hash_id"] == ancestor) & (phylolog["generation"] == generation-1)]
    if not ancestor_record.empty:
        # if ancestor is in previous generation return rank
        return ancestor_record.iloc[0]["rank"]
    # else return ids rank from previous generation
    else:
        prev_gen = phylolog.loc[(phylolog["hash_id"] == child_id) & (phylolog["generation"] == generation)]
        return prev_gen.iloc[0]["rank"]


def get_crossover_rank(ancestor, phylolog):
    if ancestor > 0:
        # print(ancestor)
        ancestor_record = phylolog.loc[(phylolog["hash_id"] == ancestor)]
        # print(ancestor_record["rank"])
        return ancestor_record.iloc[0]["rank"]
    else:
        return None


def ancestor_list(phylolog):
    return phylolog


def split_code(code):
    code = code.lstrip()
    if 'C' in code[0:1]:
        code_list = code.split('-')
        reduced_code = code.replace(code_list[0], '')
        if reduced_code[0] == '-':
            mut_only_code = reduced_code[1:]
        else:
            mut_only_code = reduced_code
        return mut_only_code
    else:
        return code


def expand_mutations_crossover(phylolog):
    phylolog["crossover_bool"] = [True if " C" in x else False for x in phylolog["mutation_code"]]
    crossover_index = [x.split('-')[0][2:] if x[0:2] == " C" else None for x in phylolog["mutation_code"]]
    # check if mutation left on end of crossover_index and remove to give crossover ancestor
    crossover_index = [[char if char.isdigit() else None for char in x] if x else None for x in crossover_index]
    phylolog["crossover_ancestor"] = [int(''.join(x)) if isinstance(x, Iterable) else None for x in crossover_index]
    # check for mutation code and save as mutation_list
    mutation_list = [x if x != " " else None for x in phylolog["mutation_code"]]
    mutation_list = [x if x and 3 <= len(x) else None for x in mutation_list]
    mutation_list = [split_code(x) if x else None for x in mutation_list]
    mutation_list = [x if x != "" else None for x in mutation_list]
    mutation_list = [x.strip() if x else x for x in mutation_list]
    # use get_ancestor_rank() to find crossover rank
    # skip empty crossovers when calling for rank
    phylolog["crossover_rank"] = phylolog.apply(
        lambda x: get_crossover_rank(x["crossover_ancestor"], phylolog), axis=1)
    phylolog["mutation_bool"] = [True if x not in [" ", "", None] else False for x in mutation_list]
    phylolog["mutation_only_code"] = mutation_list
    phylolog["mutation_type"] = [x[0] if x and 3 <= len(x) <= 5 else x for x in mutation_list]
    phylolog["mutation_token"] = [x[2] if x and 3 <= len(x) <= 5 else x for x in mutation_list]
    phylolog["mutation_operator"] = [x[3:] if x and 3 <= len(x) <= 5 else x for x in mutation_list]
    return phylolog


def add_hist_rank_stats(frame):
    # get historical data on individuals
    frame["hist_rank"] = [historical_rank(x, frame, 5, "av_rank") for x in frame["hash_id"]]
    frame["hist_change"] = frame["hash_id"].apply(historical_rank, args=[frame, 5, "av_change"])
    frame["hist_total"] = frame["hash_id"].apply(historical_rank, args=[frame, 5, "av_total"])
    return frame


def historical_rank(ind_id, frame, length, stat):
    # Create list of ancestors ranks
    ancestor_rank = []
    current = ind_id
    while len(ancestor_rank) < length:
        # get ancestor of current
        ancestor = frame.loc[frame.hash_id == current, "hash_ancestorsList"].iloc[0]
        # print(ancestor)
        if isinstance(ancestor, pd.Series):
            # print(len(ancestor))
            ancestor = ancestor[0]
        # ancestor = int(ancestor[2:-1])
        generation = frame.loc[frame.hash_id == current, "generation"].iloc[0]
        if isinstance(generation, pd.Series):
            generation = generation[0]
        # get rank of ancestor and add to list
        generation = int(generation)
        ancestor_rank.append(get_ancestor_rank(current, ancestor, generation, frame))
        # update current
        current = int(ancestor[2:-1])
    if stat == "av_rank":
        # Get average of ranks
        return np.mean(ancestor_rank)
    elif stat == "av_change":
        # Get average rank change
        ancestor_rank_change = [ancestor_rank[i]-ancestor_rank[i-1] for i in range(len(ancestor_rank))[1:]]
        return np.mean(ancestor_rank_change)
    elif stat == "total_change":
        # Get total rank change
        return ancestor_rank[-1] - ancestor_rank[0]
    else:
        return ancestor_rank


# Press the green button in the gutter to run the script.
if __name__ == '__main__':
    files = "output_files/Phylorun2/Exp1/HillClimb/*/phylolog.csv"
    li = []
    all_files = glob.glob(files)
    for file in all_files:
        phylo = load_phylo(file)
        phylo = assign_relative_fit(phylo)
        phylo = assign_relative_fit_change(phylo)
        phylo = assign_rank(phylo)
        phylo = assign_rank_change(phylo)
        phylo = ancestor_list(phylo)
        phylo = expand_mutations_crossover(phylo)
        # noinspection SpellCheckingInspection
        # file = "phylolog.csv"
        phylo = add_hist_rank_stats(phylo)
        save_phylo(phylo, file)
